import numpy as np
import torch 
torch.manual_seed(42)
from expm import expm
from utils import build_tensor

class Ising(torch.nn.Module):
    def __init__(self, chi, niter, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = 2
        self.chi = chi
        self.niter = niter
        self.dtype = dtype
        self.device = device

        B = 0.01*torch.randn(self.chi*self.D, self.chi*self.D, dtype=dtype, device=device)
        self.A = torch.nn.Parameter(B)

    def forward(self, K):
        
        # initial tensor
        K = torch.tensor(K, dtype=self.dtype, device=self.device)
        T = build_tensor(K)
        
        #isometry
        P = expm(self.A - self.A.t())
        P = P[:, :self.chi]
            
        #start CTMRG 
        C = torch.randn(self.chi, self.chi, dtype=T.dtype, device=T.device)
        E = torch.randn(self.chi, self.D, self.chi, dtype=T.dtype, device=T.device)
        for n in range(args.niter):
            A = torch.einsum('ab,eca,bdg,cdfh->efgh', (C, E, E, T)).contiguous().view(self.chi*self.D, self.chi*self.D)
            A = (A+A.t())/2.

            C = (P.t() @ A @ P)
            C = (C+C.t())/2.
            
            ET = torch.einsum('ldr,adbc->labrc', (E, T)).contiguous().view(self.chi*self.D, self.D, self.chi*self.D)
            E = torch.einsum('li,ldr,rj->idj', (P, ET, P)) 

            E = (E + E.permute(2, 1, 0))/2.

            C = C/C.norm()
            E = E/E.norm()

        Z1 = torch.einsum('ab,bcd,fd,gha,hcij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
        Z3 = torch.einsum('ab,bc,cd,da', (C,C,C,C))
        Z2 = torch.einsum('ab,bcd,de,fa,gcf,ge',(C,E,C,C,E,C))
        Z4 = torch.einsum('ab,bc,ade,cdf,eg,gf',(C,C,E,E,C,C))
        lnZ = torch.log(Z1) + torch.log(Z3) - torch.log(Z2) - torch.log(Z4)

        return lnZ 

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.44, help="beta")
    parser.add_argument("-chi", type=int, default=100, help="chi")
    parser.add_argument("-niter", type=int, default=30, help="niter")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)

    model = Ising(args.chi, args.niter, dtype=dtype, device=device)

    optimizer = torch.optim.LBFGS(model.parameters(), max_iter=10)
    params = list(model.parameters())
    params = list(filter(lambda p: p.requires_grad, params))
    nparams = sum([np.prod(p.size()) for p in params])
    print ('total nubmer of trainable parameters:', nparams)

    def closure():
        optimizer.zero_grad()
        loss = -model.forward(args.beta) # loss = -lnZ
        loss.backward()
        return loss

    for epoch in range(args.Nepochs):
        loss = optimizer.step(closure)
        print (epoch, loss.item())
