import numpy as np
import torch 
torch.manual_seed(42)
from expm import expm

class Ising(torch.nn.Module):
    def __init__(self, chi, niter, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = 2
        self.chi = chi
        self.niter = niter
        self.dtype = dtype
        self.device = device

        B = 0.1*torch.randn(niter, self.chi**2, self.chi**2, dtype=dtype, device=device)
        self.A = torch.nn.Parameter(B)

    def isometry(self, step):
        chi = min(self.D**(2**(step//2)), self.chi) 
        #print (step, self.D**(step//2+1), chi)
        A = self.A[step, :chi**2, :chi**2]
        w = expm(A - A.t())

        chi_new = min(self.chi, chi**2)
        return w[:, :chi_new].view(chi, chi, chi_new)
        
    def forward(self, K):
        
        # initial tensor
        K = torch.tensor([K], dtype=self.dtype, device=self.device)
        c = torch.sqrt(torch.cosh(K)/2.)
        s = torch.sqrt(torch.sinh(K)/2.)
        M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
        T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 

        lnZ = 0.0
        for n in range(self.niter):
            f = torch.norm(T)
            lnZ += 2**(self.niter-n)*torch.log(f)
            T = T / f

            U = self.isometry(n) # isometry for this step
            #print (T.shape , U.shape)
            
            T = torch.einsum('axob,amz,moyn,bnw->xwzy',(T, U, T, U))
            #print ('truncate done {:.3f}s'.format(time.time()-t0))
           
        trace = 0.0 
        for x in range(T.shape[0]):
            for y in range(T.shape[1]):
                trace += T[x, y, y, x]
        lnZ += torch.log(trace)
        return lnZ/2**self.niter 

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.44, help="beta")
    parser.add_argument("-chi", type=int, default=24, help="chi")
    parser.add_argument("-niter", type=int, default=30, help="niter")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)

    model = Ising(args.chi, args.niter, dtype=dtype, device=device)

    optimizer = torch.optim.LBFGS(model.parameters(), max_iter=10, tolerance_grad=0, tolerance_change=0, line_search_fn='strong_wolfe')

    params = list(model.parameters())
    params = list(filter(lambda p: p.requires_grad, params))
    nparams = sum([np.prod(p.size()) for p in params])
    print ('total nubmer of trainable parameters:', nparams)

    def closure():
        optimizer.zero_grad()
        loss = -model.forward(args.beta) # loss = -lnZ
        loss.backward()
        return loss

    for epoch in range(args.Nepochs):
        loss = optimizer.step(closure)
        print (epoch, loss.item(), model.A.grad.abs().max().item())
