import time 
import numpy as np
import torch 
torch.manual_seed(42)
torch.set_num_threads(1)

from utils import gauge, trace
import geoopt 

def renormalize(T, U):
    T = torch.einsum('axob,amz,moyn,bnw->xwzy',(T, U, T, U))

    #TU1 = torch.tensordot(T, U, dims=([0], [0]))
    #TU2 = torch.tensordot(T, U, dims=([3], [1]))
    #T = torch.tensordot(TU1, TU2, dims=([1,2,3], [1,3,0]))
    #T = T.permute(0,3,1,2).contiguous()
    return T

class Ising(torch.nn.Module):
    def __init__(self, chi, niter, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = 2
        self.chi = chi # cutoff 
        self.niter = niter
        self.dtype = dtype
        self.device = device

        self.params = torch.nn.ParameterList()
        for step in range(niter): 
            chi = min(self.D**(2**(step//2)), self.chi) 
            chi_new = min(chi**2, self.chi)

            with torch.no_grad():
                X = geoopt.ManifoldParameter(torch.randn(chi*chi, chi_new, dtype=dtype, device=device), manifold=geoopt.manifolds.Stiefel()).proj_()
            self.params.append(X)

    def get_isometry(self, step):
        return self.params[step]

    def set_isometry(self, step, U):
        self.params[step].data = U

    def init_isometry(self, T):
        '''
        initialize isometry with hotrg
        '''
        lnZ = 0.0
        for step in range(self.niter):
            f = T.abs().max()
            lnZ += 2**(-step)*torch.log(f)
            T = T / f

            #contract up-down
            vl, el = gauge(T, self.chi,'l')
            vr, er = gauge(T, self.chi,'r')

            U = vl if (el < er) else vr
            self.set_isometry(step, U.view(-1, U.shape[-1]))
            T = renormalize(T, U)
        lnZ += torch.log(trace(T))/2**args.niter
        return lnZ

    def forward(self, T):
        lnZ = 0.0
        for n in range(self.niter):
            f = T.abs().max()
            lnZ += 2**(-n)*torch.log(f)
            T = T / f

            U = self.get_isometry(n) # isometry for this step
            chi = int(np.sqrt(U.shape[0]))
            U = U.view(chi, chi, -1)
            T = renormalize(T, U)
           
        lnZ += torch.log(trace(T))/2**args.niter
        return lnZ

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.44068679350977147, help="beta")
    parser.add_argument("-chi", type=int, default=24, help="chi")
    parser.add_argument("-niter", type=int, default=30, help="niter")
    parser.add_argument("-Nepochs", type=int, default=100, help="Nepochs")
    parser.add_argument("-init", default='random', choices=['random', 'hotrg'], help="initial isometry")
    parser.add_argument("-exact", type=float, default=-0.9296953983414709, help="exact free energy")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)

    # initial tensor
    K = torch.tensor([args.beta], dtype=dtype, device=device)
    c = torch.sqrt(torch.cosh(K)/2.)
    s = torch.sqrt(torch.sinh(K)/2.)
    M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 

    model = Ising(args.chi, args.niter, dtype=dtype, device=device)
    
    if args.init == 'hotrg':
        with torch.no_grad():
            lnZ = model.init_isometry(T)
            if args.exact is not None:
                print ('initial loss with hotrg:', -lnZ.item(), (-lnZ.item()-args.exact)/np.abs(args.exact) )
            else:
                print ('initial loss with hotrg:', sweeps, -lnZ.item() )
    else:
        pass 
   
    optimizer = geoopt.optim.RiemannianAdam(model.parameters(), lr=1e-3)

    params = list(model.parameters())
    params = list(filter(lambda p: p.requires_grad, params))
    nparams = sum([np.prod(p.size()) for p in params])
    print ('total nubmer of trainable parameters:', nparams)

    def closure():
        optimizer.zero_grad()
        loss = -model.forward(T) # loss = -lnZ
        loss.backward()
        print (loss.item())
        return loss

    for epoch in range(args.Nepochs):
        loss = optimizer.step(closure)
        if args.exact is not None:
            print (epoch, loss.item(), (loss.item()-args.exact)/np.abs(args.exact) )
        else:
            print (epoch, loss.item())
